//=======================================================================
// V IDE Make Class - all the stuff needed to create
// makefiles, and run make.
//  Copyright (C) 1995-2000  Bruce E. Wampler
//
//  This program is part of the V C++ GUI Framework example programs.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  (see COPYING) along with this program; if not, write to the Free
//  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//=======================================================================

#include <v/vfilesel.h>
#include <v/vnotice.h>

#include "videapp.h"

#include "videcmdw.h"

#include "videmake.h"
//#include <direct.h>

    static int filterIndex = 0;
    static char* filter[] =
      {
        "Make*;make*;Make*.*;make*.*",
        "*",
        0
      };

//=====================>>> videMake::videMake <<<===========================
  videMake::videMake()
  {
    _stop = 0;
    cmdw = ((videApp*)theApp)->GetMsgWindow();
#ifdef V_VersionWindows
    makeTimer = new makeAuxTimer(this);	// create timer
#endif
  }

//=====================>>> videMake::~videMake <<<===========================
  videMake::~videMake()
  {
#ifdef V_VersionWindows
    makeTimer->TimerStop();
    delete makeTimer;
#endif
  }

#ifdef V_VersionWindows		// V 1.11 - new Win32 version
//=====================>>> videMake::DoMake <<<===========================
  int videMake::DoMake(videCmdWindow* parent, char* makecmd, char* what)
  {
    char cmd[maxFileNameSize];
    vOS vos;
    int isBorland = *((videApp*)theApp)->getBCCPath() != 0;

    char* mkpath = ((videApp*)theApp)->GetMkFile();
    cmdw = ((videApp*)theApp)->GetMsgWindow();

    if (!*mkpath)
      {
        vNoticeDialog note((vWindow*)cmdw);
        note.Notice("No Makefile specified. Use Select Makefile or Open Project first.");
        return 0;
      }

    if (isBorland)
	(void)fixBackSlash(mkpath,isBorland);
    else
	fixSlash(mkpath);

    cmdw->RaiseWindow();
    cmdw->SetRdOnly(0);
    if (isBorland)
	(cmdw->GetTextEd())->SetFileType(bccError);
    else
	(cmdw->GetTextEd())->SetFileType(gccError);

    // Now, change dir

    changeDir(mkpath);
    
    // build command again without what part
    strcpy(cmd,makecmd);

    if (isBorland)
	strcat(cmd, " -f");             // -f for file
    else
	strcat(cmd, " -f ");         // -n for dry run, -f for file
    strcat(cmd, mkpath);		// the name of the make file.

    if (what && *what)
      {
	strcat(cmd," ");
	strcat(cmd,what);
      }

    cmdw->AddLine(cmd);
    if (isBorland)
	cmdw->AddLine("Please wait until the complete make is done...");

    theApp->CheckEvents();
    _stop = 0;			// not stopping

    // ok, have everything ready to go in cmd...

    startMake(cmd);

    // make is now running and handled by its timer

    return 1;
  }

#else			// current X version, old Win32 version

//=====================>>> videMake::DoMake <<<===========================
  int videMake::DoMake(videCmdWindow* parent, char* makecmd, char* what)
  {
    char cmd[maxFileNameSize];
    vOS vos;
    int isBorland = *((videApp*)theApp)->getBCCPath() != 0;

    static char makeCmds[] = "makecmds.vtm" ;
    static char makeErrs[] = "makeerrs.vtm" ;

    char* mkpath = ((videApp*)theApp)->GetMkFile();
    cmdw = ((videApp*)theApp)->GetMsgWindow();

    if (!*mkpath)
      {
        vNoticeDialog note((vWindow*)cmdw);
        note.Notice("No Makefile specified. Use Select Makefile or Open Project first.");
        return 0;
      }

    if (isBorland)
	(void)fixBackSlash(mkpath,isBorland);
    else
	fixSlash(mkpath);

    cmdw->RaiseWindow();
    cmdw->SetRdOnly(0);
    if (isBorland)
	(cmdw->GetTextEd())->SetFileType(bccError);
    else
	(cmdw->GetTextEd())->SetFileType(gccError);

    strcpy(cmd,makecmd);	// just using cmd as a temp
	strcat(cmd," -f ");
    strcat(cmd,mkpath);
    if (what && *what)
      {
	strcat(cmd," ");
	strcat(cmd,what);
      }
    cmdw->AddLine(cmd);
    if (isBorland)
	cmdw->AddLine("Please wait until the complete make is done...");

    theApp->CheckEvents();
    _stop = 0;			// not stopping

    // Now, change dir

    changeDir(mkpath);
    
    // build command
    strcpy(cmd,makecmd);

    if (isBorland)
	strcat(cmd, " -f");             // -f for file
    else
	strcat(cmd, " -n -f ");         // -n for dry run, -f for file
    strcat(cmd, mkpath);		// the name of the make file.

    if (what && *what)
      {
	strcat(cmd," ");
	strcat(cmd,what);
      }

    vos.vDeleteFile(makeCmds);
    vos.vDeleteFile(makeErrs);

    if (vos.vRunProcess(cmd,makeCmds,makeErrs,1,1) != 0)
      {
	ShowErrors(cmdw,makeCmds);
	ShowErrors(cmdw,makeErrs);
	if (isBorland)
	  {
	    cmdw->AddLine("Make failed, usually because of errors.");
	  }
	else
	  {
	    cmdw->AddLine("Make failed, usually because of missing files. You may");
	    cmdw->AddLine("need to run make from a command window to see the problem.");
	  }
	cmdw->AddLine("----------");
	cmdw->SetRdOnly(1);
	cmdw->RaiseWindow();
	theApp->CheckEvents();
	return 0;
      }

    theApp->CheckEvents();
    if (_stop)
      {
	cmdw->AddLine("---- Make Aborted ----");
	cmdw->SetRdOnly(1);
	cmdw->RaiseWindow();
	theApp->CheckEvents();
	return 0;
      }

    ifstream inFile(makeCmds);

    if (!inFile)
	return 0;		// file not there

    if (!cmdw)
      {
        inFile.close();
        vos.vDeleteFile(makeCmds);
        return 0;
      }

    if (isBorland)
      {
	ShowErrors(cmdw,makeCmds);
	ShowErrors(cmdw,makeErrs);
	inFile.close();
	cmdw->AddLine("---- Make completed ----");
	cmdw->SetRdOnly(1);
	cmdw->RaiseWindow();
	theApp->CheckEvents();
	return 1;
      }

  // read and process makecmds.tmp
    const int maxBuff = 2000;
    char buff[maxBuff+2];
    char displayed[maxBuff+20];

    while (inFile.getline(buff,maxBuff))
      {
        strcpy(displayed,"> ");
        strcat(displayed,buff);
	if (!cmdw->AddLine(displayed))
	  {
	    theApp->CheckEvents();
	    vNoticeDialog note(theApp);
	    note.Notice("Make output too big -- only paritally read.");
	    break;
	  }
        char* cp;
        for (cp = buff ; *cp && *cp != '\r' && *cp != '\n' ; ++cp)
            ;
        *cp = 0;

        int retv = vos.vRunProcess(buff,0,makeErrs,1,1);
        
        ShowErrors(cmdw,makeErrs);

        theApp->CheckEvents();
        if (_stop)
          {
            inFile.close();
            vos.vDeleteFile(makeCmds);
	    cmdw->AddLine("---- Make Aborted ----");
            cmdw->SetRdOnly(1);
	    cmdw->RaiseWindow();
	    theApp->CheckEvents();
            return 0;
          }

        if (retv != 0)
          {
            inFile.close();
            vos.vDeleteFile(makeCmds);
            cmdw->AddLine("Make failed. Right click error line to open file.");
            cmdw->SetRdOnly(1);
	    cmdw->RaiseWindow();
	    theApp->CheckEvents();
            return 0;
          }
        }

    inFile.close();
    vos.vDeleteFile(makeCmds);
    cmdw->AddLine("--------");
    cmdw->SetRdOnly(1);
    cmdw->RaiseWindow();
    theApp->CheckEvents();
    return 1;
  }
#endif

//=====================>>> videMake::ShowErrors <<<=========================
  void videMake::ShowErrors(videCmdWindow* cmdw, char* errName)
  {
    const int maxBuff = 1000;
    char buff[maxBuff+2];
    char displayed[maxBuff+20];
    vOS vos;

    ifstream errFile(errName);

    if (!errFile)
        return;

    while (errFile.getline(buff,maxBuff))
      {
        strcpy(displayed,"  ! ");	// IMPORTANT!  videCmdWindow::GotoErrorLine
        strcat(displayed,buff);         //	REQUIRES this exact format!!!
        cmdw->AddLine(displayed);
        theApp->CheckEvents();
        if (_stop)
          {
	    cmdw->AddLine("---- Make Aborted ----");
	    vos.vDeleteFile(errName);
            return;
          }

      }
    errFile.close();
    vos.vDeleteFile(errName);
  }

//=====================>>> videMake::CheckMake <<<===========================
  int videMake::CheckMake(char* makecmd,char* what)
  {
    // Check if make would do anything...
    char cmd[maxFileNameSize];
    vOS vos;

    static char makeCmds[] = "makecmds.vtm" ;
    static char makeErrs[] = "makeerrs.vtm" ;

    char* mkpath = ((videApp*)theApp)->GetMkFile();

    videCmdWindow* cmdw = ((videApp*)theApp)->GetMsgWindow();

    if (*((videApp*)theApp)->getBCCPath() != 0)
	return 0;				// just fail if BCC

    if (!*mkpath)
      {
	return 0;
      }

    // Now, change directory
    strcpy(cmd,mkpath);

    int ix = strlen(cmd);
    while (ix > 0)
      {
        if (cmd[ix] == '\\' || cmd[ix] == '/')
          {
            cmd[ix] = 0;
            break;
          }
        --ix;
      }
    char makeFile[100];
    strcpy(makeFile, &cmd[ix+1]);	// copy the makefile name
    vos.vChDir(cmd);			// change dir

    strcpy(cmd, makecmd);		// make or make clean passed from caller
    strcat(cmd, " -n -f ");             // -n for dry run, -f for file
    strcat(cmd, makeFile);		// the name of the make file.

    if (what && *what)
      {
	strcat(cmd, " ");
	strcat(cmd, what);
      }
    if (vos.vRunProcess(cmd,makeCmds,makeErrs,1,1) != 0)
      {
	vos.vDeleteFile(makeErrs);
	vos.vDeleteFile(makeCmds);
     	return 0;
      }
    // OK, then perhaps "nothing to make" issued.
    ifstream inFile(makeCmds);			// open the makecmds file

    int retv = 1;
    char buff[256];
    if (inFile.getline(buff,155))
      {
	if (strstr(buff, "Nothing to be done") != 0)
	    retv = 0;
      }
    inFile.close();

    vos.vDeleteFile(makeErrs);
    vos.vDeleteFile(makeCmds);
    return retv;
  }

// The following code is used for the new way to run a make in Win32
#ifdef V_VersionWindows		// V 1.11 - new Win32 version
//=====================>>> videMake::DoMake <<<===========================
  int videMake::startMake(char* cmd)
  {
     /*
      * The idea is to run the make in the "background". The make
      * process will run until it is done, sending all stderr and stdout
      * to a pipe. We will read that pipe using a timer. We will also
      * check to see if the make process has completed, and stop the
      * timer when that happens. We will also moniter the _stop
      * variable to see if the user aborted the make.
      */

    strcpy(makeOutLine,"  ! ");		// set up output line
    makeOutNext = 4;

    SECURITY_ATTRIBUTES saAttr; 
 
    /* Set the bInheritHandle flag so pipe handles are inherited. */ 

    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;
 
    /* 
     * The steps for redirecting child's STDOUT: 
     *     1.  Save current STDOUT, to be restored later. 
     *     2.  Create anonymous pipe to be STDOUT for child. 
     *     3.  Set STDOUT of parent to be write handle of pipe, so 
     *         it is inherited by child.
     */ 
 
    /* Save the handle to the current STDOUT */ 

    hSaveStdout = ::GetStdHandle(STD_OUTPUT_HANDLE);
 
    /* Create a pipe for the child's STDOUT. */ 
 
    if (! ::CreatePipe(&hChildStdoutRd, &hChildStdoutWr, &saAttr, 0))
        return 0;	// ErrorExit("Stdout pipe creation failed\n");
 
    /* Set a write handle to the pipe to be STDOUT. */ 

    if (! ::SetStdHandle(STD_OUTPUT_HANDLE, hChildStdoutWr))
        return 0;	// ErrorExit("Redirecting STDOUT failed");

   /* 
     * The steps for redirecting child's STDIN: 
     *     1.  Save current STDIN, to be restored later. 
     *     2.  Create anonymous pipe to be STDIN for child. 
     *     3.  Set STDIN of parent to be read handle of pipe, so 
     *         it is inherited by child. 
     *     4.  Create a noninheritable duplicate of write handle, 
     *         and close the inheritable write handle. 
     */ 
 
    /* Save the handle to the current STDIN. */ 

    hSaveStdin = GetStdHandle(STD_INPUT_HANDLE);
 
    /* Create a pipe for the child's STDIN. */ 
 
    if (! CreatePipe(&hChildStdinRd, &hChildStdinWr, &saAttr, 0)) 
        return 0;	// ErrorExit("Stdin pipe creation failed\n");

    /* Set a read handle to the pipe to be STDIN. */ 
 
    if (! SetStdHandle(STD_INPUT_HANDLE, hChildStdinRd)) 
        return 0;	// ErrorExit("Redirecting Stdin failed"); 
 
    /* Duplicate the write handle to the pipe so it is not inherited. */
 
    BOOL fSuccess = DuplicateHandle(GetCurrentProcess(), hChildStdinWr, 
        GetCurrentProcess(), &hChildStdinWrDup, 0, 
        TRUE,       /* is inherited */ 
        DUPLICATE_SAME_ACCESS); 

    if (! fSuccess) 
        return 0;	// ErrorExit("DuplicateHandle failed"); 
 
    CloseHandle(hChildStdinWr);  
 

    /* Now create the child process. */

    static PROCESS_INFORMATION piProcInfo;
    STARTUPINFO siStartInfo; 

    /* Set up members of STARTUPINFO structure. */ 

    ::ZeroMemory(&siStartInfo, sizeof(siStartInfo));
    siStartInfo.cb = sizeof(STARTUPINFO); 
    siStartInfo.lpReserved = NULL; 
    siStartInfo.lpReserved2 = NULL; 
    siStartInfo.cbReserved2 = 0; 
    siStartInfo.lpDesktop = NULL;
    siStartInfo.wShowWindow = SW_MINIMIZE;
//    siStartInfo.wShowWindow = SW_SHOWNORMAL;
    siStartInfo.dwFlags = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
    siStartInfo.hStdOutput = hChildStdoutWr;
    siStartInfo.hStdError = hChildStdoutWr;
    siStartInfo.hStdInput = hChildStdinRd;

    /* Create the child process. */
    OSVERSIONINFO osver;
    osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    int createResult;

    ::GetVersionEx(&osver);
    if (osver.dwPlatformId == VER_PLATFORM_WIN32_NT)
      {
	createResult = ::CreateProcess(NULL,
		cmd,
		NULL,          // process security attributes
		NULL,          // primary thread security attributes
		TRUE,          // handles are inherited
		CREATE_NEW_PROCESS_GROUP,   /* creation flags                     */
//	 CREATE_NEW_CONSOLE, //| CREATE_NEW_PROCESS_GROUP,
		NULL,          /* use parent's environment           */
		NULL,          /* use parent's current directory     */ 

		&siStartInfo,  /* STARTUPINFO pointer                */ 
		&piProcInfo);  /* receives PROCESS_INFORMATION       */
      }
    else		// Win9x version
      {
	createResult = ::CreateProcess(NULL,
		cmd,
		NULL,          /* process security attributes        */ 
		NULL,          /* primary thread security attributes */ 
		TRUE,          /* handles are inherited              */
		0,             /* creation flags                     */
// maybe:	CREATE_NEW_CONSOLE,
		NULL,          /* use parent's environment           */
		NULL,          /* use parent's current directory     */ 

		&siStartInfo,  /* STARTUPINFO pointer                */ 
		&piProcInfo);  /* receives PROCESS_INFORMATION       */
      }

    if (!createResult)
      {
	static DWORD errorCode = ::GetLastError();
	if (errorCode != 0)
	  {
	    cmdw->AddLine("CreateProcess Fails....");
	    cmdw->AddLine(cmd);
	  }

	return 0;	// create fails
      }


    // Now, start the timer to monitor make progress
    makeTimer->TimerSet(300);		// update 3 times a second

    *makeCmd = 0;
    makeCmdPtr = makeCmd;		// build command in makeCmd

    makeProcessId = piProcInfo.dwProcessId;	// save these for termination
    makeProcessHandle = piProcInfo.hProcess;
    makeProcessThread = piProcInfo.hThread;	// save these for termination

    /* After process creation, restore the saved STDIN and STDOUT. */

    if (! SetStdHandle(STD_OUTPUT_HANDLE, hSaveStdout))
        return 0;	// ErrorExit("Re-redirecting Stdout failed\n");
    if (! SetStdHandle(STD_OUTPUT_HANDLE, hSaveStdout))
        return 0;	// ErrorExit("Re-redirecting Stdout failed\n");

    return 1;
    
  }

//==================>>> videMake::checkMakeProgress <<<======================
  void videMake::checkMakeProgress()
  {
    // this will try to echo the output of the Make, and terminate the
    // make on _stop or when done.
    int bytesRead;

    char chrin[10] = "";

    // First, read any characters output
    for ( ; ; )		// read all debugger output available
      {
	// first, see if any output is available...
	bytesRead = readMake(chrin);
	if (bytesRead == 0)		// nothing to read!
	    break;
	if (*chrin == '\r')		// ignore these
	     continue;
	else if (*chrin == '\n')		// end of line
	  {
	    makeOutLine[makeOutNext] = 0;
	    if (strstr(makeOutLine,": ") != 0)	// error line
		cmdw->AddLine(makeOutLine);
	    else				// normal line
	      {
		makeOutLine[2] = '>';
		cmdw->AddLine(&makeOutLine[2]);
	      }
	    strcpy(makeOutLine,"  ! ");
	    makeOutNext = 4;
	  }
	else
	  {
	    makeOutLine[makeOutNext++] = *chrin;
	    if (makeOutNext >= makeOutMax)
	      {
		makeOutLine[makeOutNext] = 0;
		if (strstr(makeOutLine,": ") != 0)	// error line
		    cmdw->AddLine(makeOutLine);
		else				// normal line
		  {
		    makeOutLine[2] = '>';
		    cmdw->AddLine(&makeOutLine[2]);
		  }
		strcpy(makeOutLine,"  ! ");
		makeOutNext = 4;
	      }
	  }
      }

    // Second, check for stop
    theApp->CheckEvents();
    if (_stop)
      {
	_stop = 0;
	makeTimer->TimerStop();
	cmdw->AddLine("---- Make Aborted ----");
	cmdw->SetRdOnly(1);
	cmdw->RaiseWindow();
	theApp->CheckEvents();
	::TerminateProcess(makeProcessHandle,99);
	::CloseHandle(makeProcessThread);
	::CloseHandle(makeProcessHandle);
	return;
      }

    // Third, see if make is finished
    DWORD status = ::WaitForSingleObject(makeProcessHandle,(DWORD)0);
    if (status == WAIT_OBJECT_0)	// signaled, means done
      {
	makeTimer->TimerStop();
	DWORD retv = 0;
	::GetExitCodeProcess(makeProcessHandle,&retv);
	if (retv != 0)
	    cmdw->AddLine("---- Make had errors. Right click error line to open file.");
	else
	    cmdw->AddLine("---- Make finished ----");
	cmdw->SetRdOnly(1);
	cmdw->RaiseWindow();
	theApp->CheckEvents();
	::CloseHandle(makeProcessThread);
	::CloseHandle(makeProcessHandle);
      }

  }

//====================>>> vdbCmdWindow::readMake <<<====================
  int videMake::readMake(char* chrin)
  {
    // read a char from make if there, otherwise return 0

    DWORD bytesRead;
    DWORD bytesAvail;

    if (!PeekNamedPipe(hChildStdoutRd,
	chrin, (DWORD)1, &bytesRead, &bytesAvail, 0))
      {
	return 0;
      }
    if (bytesAvail == 0)	// no output avail, so return
	return 0;

    if (!ReadFile(hChildStdoutRd,	// now really read it
	chrin, (DWORD)1, &bytesRead, 0) )
    	return 0;
    return bytesRead;
  }

//=======================>>> myAuxTimer::TimerTick <<<=======================
  void makeAuxTimer::TimerTick()
  {
    makeIt->checkMakeProgress();
  }
#endif

